// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AI64SSC750K/driver/device.c $
// $Rev: 53319 $
// $Date: 2023-07-05 16:01:02 -0500 (Wed, 05 Jul 2023) $

// 18AI64SSC750K: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type

	{ "18AI64SSC750K",	0x10B5, 0x9056, 0x10B5, 0x3570,	GSC_DEV_TYPE_18AI64SSC750K	},
	{ NULL }
};



//*****************************************************************************
static void _b33_compute(dev_data_t* dev)
{
	if (dev->cache.gsc_bcfgr_32 & D17)
		dev->cache.b33_func	= AI64SSC750K_B33_FUNC_CLK_IO;
	else
		dev->cache.b33_func	= AI64SSC750K_B33_FUNC_IN_RET;
}



//*****************************************************************************
static void _channels_compute(dev_data_t* dev)
{
	if (dev->cache.gsc_bcfgr_32 & D12)
		dev->cache.channel_qty	= 32;
	else
		dev->cache.channel_qty	= 64;

	dev->cache.channels_max	= 64;
}



//*****************************************************************************
static void _input_bandwidth_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_bcfgr_32 & 0x18000)
	{
		default:
		case 0x00000:	dev->cache.input_bandwidth	= 340000L;	break;
		case 0x08000:	dev->cache.input_bandwidth	=  70000L;	break;
		case 0x10000:	dev->cache.input_bandwidth	=	   0L;	break;
		case 0x18000:	dev->cache.input_bandwidth	=	   0L;	break;
	}
}



//*****************************************************************************
static void _master_clock_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_bcfgr_32 & 0x6000)
	{
		default:
		case 0x2000:	// reserved as of 20150119
		case 0x4000:	// reserved as of 20150119
		case 0x6000:	// reserved as of 20150119
		case 0x0000:	dev->cache.master_clock	= 60000000L;	break;
	}
}



//*****************************************************************************
static void _v_range_compute(dev_data_t* dev)
{
	if (dev->cache.gsc_bcfgr_32 & D18)
		dev->cache.v_range_high	= 0;
	else
		dev->cache.v_range_high	= 1;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AI64SSC750K_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_acar_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_ACAR));
		dev->vaddr.gsc_asiocr_32	= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_ASIOCR));
		dev->vaddr.gsc_bcfgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_BCFGR));
		dev->vaddr.gsc_bctlr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_BCTLR));
		dev->vaddr.gsc_bufsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_BUFSR));
		dev->vaddr.gsc_bursr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_BURSR));
		dev->vaddr.gsc_ibcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_IBCR));
		dev->vaddr.gsc_icr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_ICR));
		dev->vaddr.gsc_idbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_IDBR));
		dev->vaddr.gsc_llhr00_16	= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_LLHR00));
		dev->vaddr.gsc_ragr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_RAGR));
		dev->vaddr.gsc_rbgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_RBGR));
		dev->vaddr.gsc_smlwr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_SMLWR));
		dev->vaddr.gsc_smuwr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_SMUWR));
		dev->vaddr.gsc_sscr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AI64SSC750K_GSC_SSCR));

		// Data cache initialization.
		dev->cache.gsc_bcfgr_32		= os_reg_mem_rx_u32(NULL, dev->vaddr.gsc_bcfgr_32);

		dev->cache.autocal_ms		= 4000;
		dev->cache.fifo_size		= _256K;
		dev->cache.fsamp_max		= 750000L;
		dev->cache.fsamp_min		= 1L;
		dev->cache.initialize_ms	= 3;
		dev->cache.rate_gen_qty		= 2;

		_b33_compute(dev);
		_channels_compute(dev);
		_input_bandwidth_compute(dev);
		_master_clock_compute(dev);
		_v_range_compute(dev);

		dev->cache.nrate_max		= 0xFFFF;
		dev->cache.nrate_min		= dev->cache.master_clock / dev->cache.fsamp_max;

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_READ
			| GSC_DMA_CAP_DMDMA_READ;
		ret	= gsc_dma_create(dev, dma, dma);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


